#!/usr/bin/env python
"""XXX
"""

from __future__ import print_function
from ZODB import DB
from ZODB.POSException import ConflictError
import transaction
from persistent import Persistent
from random import randint

from golang import func, defer, select, default
from golang import sync, context


# PInt is persistent integer.
class PInt(Persistent):
    def __init__(self, i):
        self.i = i

from wendelin.lib.testing import TestDB_ZEO


@func
def main():
    tdb = TestDB_ZEO('<zeo>')
    tdb.setup()
    defer(tdb.teardown)

    def dbopen():
        zstor = tdb.getZODBStorage()
        db = DB(zstor)
        return db

    # init initializes the database with two integer objects - obj1/obj2 that are set to 0.
    @func
    def init():
        db = dbopen()
        defer(db.close)

        transaction.begin()
        zconn = db.open()

        root = zconn.root()
        root['obj1'] = PInt(0)
        root['obj2'] = PInt(0)

        transaction.commit()
        zconn.close()


    # T is a worker that accesses obj1/obj2 in a loop and verifies
    # `obj1.i == obj2.i` invariant.
    #
    # access to obj1 is organized to always trigger loading from zstor.
    # access to obj2 goes through zconn cache and so verifies whether the cache is not stale.
    #
    # Once in a while T tries to modify obj{1,2}.i maintaining the invariant as
    # test source of changes for other workers.
    @func
    def T(ctx, name, N):
        db = dbopen()
        defer(db.close)

        @func
        def t1():
            transaction.begin()
            zconn = db.open()
            defer(zconn.close)

            root = zconn.root()
            obj1 = root['obj1']
            obj2 = root['obj2']

            # obj1 - reload it from zstor
            # obj2 - get it from zconn cache
            obj1._p_invalidate()

            # both objects must have the same values
            i1 = obj1.i
            i2 = obj2.i
            if i1 != i2:
                #print('FAIL')
                raise AssertionError("T%s: obj1.i (%d)  !=  obj2.i (%d)" % (name, i1, i2))

            # change objects once in a while
            if randint(0,4) == 0:
                #print("T%s: modify" % name)
                obj1.i += 1
                obj2.i += 1

            try:
                transaction.commit()
            except ConflictError:
                #print('conflict -> ignore')
                transaction.abort()

        for i in range(N):
            if ready(ctx.done()):
                break
            #print('T%s.%d' % (name, i))
            t1()


    # run 8 T workers concurrently. As of 20200123, likely due to race conditions
    # in ZEO, it triggers the bug where T sees stale obj2 with obj1.i != obj2.i
    init()

    N = 100000
    wg = sync.WorkGroup(context.background())
    for x in range(8):
        wg.go(T, x, N)
    wg.wait()

    print('OK')


# ready returns whether channel ch is ready.
def ready(ch):
    _, _rx = select(
        default,        # 0
        ch.recv,        # 1
    )
    if _ == 0:
        return False
    return True

if __name__ == '__main__':
    main()
